import cocotb
from cocotb.clock import Clock
from cocotb.triggers import ClockCycles, RisingEdge
from {{name}} import {{top_entity}}


CLK_FREQ = 312.5e6


async def init(dut):
    """ Ideally this would be a test fixture hidden away in lib/
    """
    cocotb.start_soon(Clock(dut.Clk, round(1e9 / CLK_FREQ), units='ns').start())
    dut.Reset.value = 1
    await ClockCycles(dut.Clk, 10)
    await RisingEdge(dut.Clk)
    dut.Reset.value = 0

    return {{top_entity}}(dut)


@cocotb.test(timeout_time=100, timeout_unit='us')
async def test_one(dut):
    {{name}}_inst = await init(dut)

    {{name}}_inst.reg0 = 0xDEADBEEF
    {{name}}_inst.reg1 = 0xCAFEBABE
    await {{name}}_inst.apply()  # async wait for the SPI bus

    await ClockCycles(dut.Clk, 100)



from .{{name.lower()}} import {{top_entity}}
try:
    from .applet import {{name}}Frame
except ImportError:
    # the plugin should still work without all the GUI dependencies
    # note that we can't defer imports in a plugin
    pass


class Plugin:
    def get_name(self, _id):
        """ An identifiable name for this slot based on it's id
        """
        if len(_id.split(':')) > 2:
            return f"{{name}} {_id.split(':')[-1].title()}"

        return f"{{name}}"

    def attach(self, moku, manifest):
        """ Create an instance of this slot, attached to the given slot and moku
        """
        return {{top_entity}}(moku, manifest['slot'])

    def applet_frame(self):
        """ Create a DefaultFrame instance for this slot
        """
        return {{name}}Frame



# Copyright (c) 2014 Potential Ventures Ltd
# Copyright 2019 by Ben Coughlan <ben@liquidinstruments.com>
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#     * Redistributions of source code must retain the above copyright
#       notice, this list of conditions and the following disclaimer.
#     * Redistributions in binary form must reproduce the above copyright
#       notice, this list of conditions and the following disclaimer in the
#       documentation and/or other materials provided with the distribution.
#     * Neither the name of Potential Ventures Ltd,
#       SolarFlare Communications Inc nor the
#       names of its contributors may be used to endorse or promote products
#       derived from this software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND
# ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL POTENTIAL VENTURES LTD BE LIABLE FOR ANY
# DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
# (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
# LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
# ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
# (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
# SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

"""Drivers for Advanced Microcontroller Bus Architecture."""

import cocotb
from cocotb.triggers import RisingEdge, ReadOnly, Lock
from cocotb_bus.drivers import BusDriver
from cocotb.queue import Queue
from cocotb.binary import BinaryValue


class AXIProtocolError(Exception):
    pass


class AXI4LiteMaster(BusDriver):
    """AXI4-Lite Master.

    TODO: Kill all pending transactions if reset is asserted.
    """

    _signals = ["AWVALID", "AWADDR", "AWREADY",        # Write address channel
                "WVALID", "WREADY", "WDATA", "WSTRB",  # Write data channel
                "BVALID", "BREADY", "BRESP",           # Write response channel
                "ARVALID", "ARADDR", "ARREADY",        # Read address channel
                "RVALID", "RREADY", "RRESP", "RDATA"]  # Read data channel

    def __init__(self, entity, name, clock, **kwargs):
        BusDriver.__init__(self, entity, name, clock, **kwargs)

        # Drive some sensible defaults (setimmediatevalue to avoid x asserts)
        self.bus.AWVALID.setimmediatevalue(0)
        self.bus.WVALID.setimmediatevalue(0)
        self.bus.ARVALID.setimmediatevalue(0)
        self.bus.BREADY.setimmediatevalue(1)
        self.bus.RREADY.setimmediatevalue(1)

        # Mutex for each channel that we master to prevent contention
        self.write_address_busy = Lock("%s_wabusy" % name)
        self.read_address_busy = Lock("%s_rabusy" % name)
        self.write_data_busy = Lock("%s_wbusy" % name)

    @cocotb.coroutine
    def _send_write_address(self, address, delay=0):
        """
        Send the write address, with optional delay (in clocks)
        """
        yield self.write_address_busy.acquire()
        for cycle in range(delay):
            yield RisingEdge(self.clock)

        self.bus.AWADDR.value = address
        self.bus.AWVALID.value = 1
        yield RisingEdge(self.clock)

        while True:
            if self.bus.AWREADY.value:
                break
            yield RisingEdge(self.clock)

        self.bus.AWVALID.value = 0
        self.write_address_busy.release()

    @cocotb.coroutine
    def _send_write_data(self, data, delay=0, byte_enable=0xF):
        """Send the write address, with optional delay (in clocks)."""
        yield self.write_data_busy.acquire()
        for cycle in range(delay):
            yield RisingEdge(self.clock)

        self.bus.WDATA.value = data
        self.bus.WVALID.value = 1
        self.bus.WSTRB.value = byte_enable
        yield RisingEdge(self.clock)

        while True:
            if self.bus.WREADY.value:
                break
            yield RisingEdge(self.clock)

        self.bus.WVALID.value = 0
        self.write_data_busy.release()

    @cocotb.coroutine
    def write(self, address, value, byte_enable=0xf, address_latency=0,
              data_latency=0, sync=True):
        """Write a value to an address.

        Args:
            address (int): The address to write to.
            value (int): The data value to write.
            byte_enable (int, optional): Which bytes in value to actually write.
                Default is to write all bytes.
            address_latency (int, optional): Delay before setting the address (in clock cycles).
                Default is no delay.
            data_latency (int, optional): Delay before setting the data value (in clock cycles).
                Default is no delay.
            sync (bool, optional): Wait for rising edge on clock initially.
                Defaults to True.

        Returns:
            BinaryValue: The write response value.

        Raises:
            AXIProtocolError: If write response from AXI is not ``OKAY``.
        """
        if sync:
            yield RisingEdge(self.clock)

        c_addr = cocotb.start_soon(self._send_write_address(address,
                                                            delay=address_latency))
        c_data = cocotb.start_soon(self._send_write_data(value,
                                                         byte_enable=byte_enable,
                                                         delay=data_latency))

        if not c_addr.done():
            yield c_addr.join()
        if not c_data.done():
            yield c_data.join()

        # Wait for the response
        while True:
            yield ReadOnly()
            if self.bus.BVALID.value and self.bus.BREADY.value:
                result = self.bus.BRESP.value
                break
            yield RisingEdge(self.clock)

        yield RisingEdge(self.clock)

        if int(result):
            raise AXIProtocolError("Write to address 0x%08x failed with BRESP: %d"
                                   % (address, int(result)))

        return result

    @cocotb.coroutine
    def read(self, address, sync=True):
        """Read from an address.

        Args:
            address (int): The address to read from.
            sync (bool, optional): Wait for rising edge on clock initially.
                Defaults to True.

        Returns:
            BinaryValue: The read data value.

        Raises:
            AXIProtocolError: If read response from AXI is not ``OKAY``.
        """
        if sync:
            yield RisingEdge(self.clock)

        self.bus.ARADDR.value = address
        self.bus.ARVALID.value = 1

        while True:
            yield ReadOnly()
            if self.bus.ARREADY.value:
                break
            yield RisingEdge(self.clock)

        yield RisingEdge(self.clock)
        self.bus.ARVALID.value = 0

        while True:
            yield ReadOnly()
            if self.bus.RVALID.value and self.bus.RREADY.value:
                data = self.bus.RDATA.value
                result = self.bus.RRESP.value
                break
            yield RisingEdge(self.clock)

        if int(result):
            raise AXIProtocolError("Read address 0x%08x failed with RRESP: %d" %
                                   (address, int(result)))

        return data

    def __len__(self):
        return 2**len(self.bus.ARADDR)


@cocotb.coroutine
def handshake(clock, drive, listen, delay=-1):
    """ Wait for a valid handshake on the supplied AXI signals

    `drive` and `listen` are the valid and ready signals for a given AXI
    channel depending on which direction this handshake is occuring. This
    coroutine will drive `drive` and wait for `listen`.

    `delay` specifies the clock cycles to wait before asserting drive
    after the master asserts listen.
        delay < 0: drive is asserted before listen.
        delay = 0: drive asserted combinatorially with listen.
        delay > 0: wait 'delay' clock cycles after listen to assert drive.

        If delay is a generator it will be next()'d
    """
    try:
        delay = next(delay)
    except TypeError:
        pass

    drive.value = int(delay < 0)

    # check if listen is already asserted
    if not listen.value:
        yield RisingEdge(listen)

    # artificial delay
    for i in range(delay):
        yield RisingEdge(clock)

    # handshake - either drive or listen may already be asserted
    drive.value = 1
    yield RisingEdge(clock)
    while not listen.value:
        yield RisingEdge(clock)
    drive.value = 0


class AXI4Master(BusDriver):
    """AXI4 Master
    """
    _signals = [
        "ARID", "ARADDR", "ARLEN", "ARSIZE", "ARBURST", "ARVALID", "ARREADY",
        "AWID", "AWADDR", "AWLEN", "AWSIZE", "AWBURST", "AWVALID", "AWREADY",
        "WID", "WDATA", "WSTRB", "WLAST", "WVALID", "WREADY",
        "RID", "RDATA", "RRESP", "RLAST", "RVALID", "RREADY",
        "BID", "BRESP", "BVALID", "BREADY",
    ]

    # Not currently supported by this driver
    _optional_signals = [
        "ARLOCK", "ARCACHE", "ARPROT", "ARQOS", "ARREGION", "ARUSER",
        "AWLOCK", "AWCACHE", "ARPROT", "AWQOS", "AWREGION", "AWUSER",
        "WUSER", "RUSER", "BUSER",
    ]

    def __init__(self, entity, name, clock, max_burst=256, max_threads=1, **kwargs):
        BusDriver.__init__(self, entity, name, clock, **kwargs)
        self.clock = clock

        self.bus.AWVALID.value = 0
        self.bus.ARVALID.value = 0
        self.bus.WVALID.value = 0
        self.bus.WLAST.value = 0
        self.bus.RREADY.value = 0
        self.bus.BREADY.value = 0

        self._bus_width = (
            abs(self.bus.WDATA._range[0] - self.bus.WDATA._range[1]) + 1) // 8

        assert self._bus_width in [2**n for n in range(8)], \
            "Unsupported Bus Width"

        self.max_burst = max_burst

        self._write_threads = Queue()
        self._read_threads = Queue()

        for i in range(max_threads):
            self._write_threads.put_nowait(i)
            self._read_threads.put_nowait(i)

        self.aw_lock = Lock("%s_aw_lock" % name)
        self.ar_lock = Lock("%s_ar_lock" % name)
        self.w_lock = Lock("%s_w_lock" % name)

    async def _write_address(self, wid, addr, length, size=0b011, burst=0b01):
        await self.aw_lock.acquire()
        self.bus.AWADDR.value = addr
        self.bus.AWLEN.value = length - 1
        self.bus.AWID.value = wid
        self.bus.AWSIZE.value = size
        self.bus.AWBURST.value = burst
        await handshake(self.clock, self.bus.AWVALID, self.bus.AWREADY)
        self.aw_lock.release()

    async def _write_data(self, data, wid):
        assert len(data) <= self.max_burst * self._bus_width
        assert len(data) % self._bus_width == 0

        await self.w_lock.acquire()
        self.bus.WID.value = wid
        for i in range(0, len(data), self._bus_width):
            # self.bus.WDATA.value.big_endian = False
            self.bus.WDATA.value = BinaryValue(data[i:i + self._bus_width], bigEndian=False)
            self.bus.WSTRB.value = 0b011111111  # TODO
            self.bus.WLAST.value = int((i + 1) * self._bus_width == len(data))
            await handshake(self.clock, self.bus.WVALID, self.bus.WREADY)
        self.bus.WLAST.value = 0
        self.w_lock.release()

    async def _write_response(self, wid):
        await handshake(self.clock, self.bus.BREADY, self.bus.BVALID)

    async def _write_thread(self, addr, data):
        wid = await self._write_threads.get()
        wdata = cocotb.start_soon(self._write_data(data, wid))
        waddr = cocotb.start_soon(self._write_address(wid, addr, len(data) // self._bus_width))
        await wdata
        await waddr
        await self._write_response(wid)
        self._write_threads.put_nowait(wid)

    async def write(self, addr, data):
        """ Write bytes to the AXI interface
        Args:
            addr: start address
            data: bytearray
        """
        tasks = []
        length = len(data)
        i = 0

        while i < len(data):
            length = min(self.max_burst * self._bus_width, len(data) - i)

            tasks.append(cocotb.start_soon(
                self._write_thread(
                    addr=i + addr,
                    data=data[i:i + length]
                )))
            i += length

        for task in tasks:
            await task

    async def _read_address(self, rid, addr, length, size=0b011, burst=0b01):
        await self.ar_lock.acquire()
        self.bus.ARADDR.value = addr
        self.bus.ARLEN.value = max((length // self._bus_width) - 1, 0)
        self.bus.ARID.value = rid
        self.bus.ARSIZE.value = size  # TODO
        self.bus.ARBURST.value = burst
        await handshake(self.clock, self.bus.ARVALID, self.bus.ARREADY)
        self.ar_lock.release()

    async def _read_data(self, rid):
        data = bytearray()
        while True:
            await handshake(self.clock, self.bus.RREADY, self.bus.RVALID)

            if self.bus.RID.value != rid:
                continue

            assert self.bus.RRESP.value == 0, 'RRESP returned error'

            val = self.bus.RDATA.value
            val.big_endian = False
            data += val.buff

            if self.bus.RLAST.value:
                return data

    async def _read_thread(self, addr, length):
        rid = await self._read_threads.get()
        raddr = cocotb.start_soon(self._read_address(rid, addr, length))
        rdata = cocotb.start_soon(self._read_data(rid))
        await raddr
        data = await rdata
        self._read_threads.put_nowait(rid)
        return data

    async def read(self, addr, length):
        """ Read bytes from the AXI interface
        Args:
            addr: start address
            length: in bytes
        """
        tasks = []
        data = bytearray()

        # split long read into many transactions.  These may run as parallel
        # threads if max_threads > 1
        for i in range(0, length, self.max_burst * self._bus_width):
            tasks.append(cocotb.start_soon(
                self._read_thread(
                    addr=i + addr,
                    length=min(self.max_burst * self._bus_width, length - i)
                )))

        for task in tasks:
            # assuming they all complete in order
            data += await task

        return data


class AXI4Slave(BusDriver):
    """AXI4 Slave
    """
    _signals = [
        "ARID", "ARADDR", "ARLEN", "ARSIZE", "ARBURST", "ARVALID", "ARREADY",
        "AWID", "AWADDR", "AWLEN", "AWSIZE", "AWBURST", "AWVALID", "AWREADY",
        "WID", "WDATA", "WSTRB", "WLAST", "WVALID", "WREADY",
        "RID", "RDATA", "RRESP", "RLAST", "RVALID", "RREADY",
        "BID", "BRESP", "BVALID", "BREADY",
    ]

    # Not currently supported by this driver
    _optional_signals = [
        "ARLOCK", "ARCACHE", "ARPROT", "ARQOS", "ARREGION", "ARUSER",
        "AWLOCK", "AWCACHE", "ARPROT", "AWQOS", "AWREGION", "AWUSER",
        "WUSER", "RUSER", "BUSER",
    ]

    def __init__(self, entity, name, clock, max_threads=1,
                 awready_delay=-1, wready_delay=1, bresp_delay=1,
                 arready_delay=-1, rresp_delay=1, **kwargs):
        BusDriver.__init__(self, entity, name, clock, **kwargs)
        self.clock = clock

        self.bus.RVALID.value = 0
        self.bus.RLAST.value = 0
        self.bus.BVALID.value = 0
        self.bus.BID.value = 0

        self.awready_delay = awready_delay
        self.wready_delay = wready_delay
        self.arready_delay = arready_delay
        assert bresp_delay > 0, "bresp_delay < 1 not supported"
        self.bresp_delay = bresp_delay
        assert rresp_delay > 0, "rresp_delay < 1 not supported"
        self.rresp_delay = rresp_delay

        self._bus_width = (
            abs(self.bus.WDATA._range[0] - self.bus.WDATA._range[1]) + 1) // 8

        assert self._bus_width in [2**n for n in range(8)], \
            "Unsupported Bus Width"

        self.max_threads = max_threads
        self._write_threads = []
        self._read_threads = []

        self.bresp_lock = Lock("%s_bresp_lock" % name)
        self.rresp_lock = Lock("%s_rresp_lock" % name)

        cocotb.start_soon(self._read_transaction())
        cocotb.start_soon(self._write_transaction())

    @cocotb.coroutine
    def _write_address(self):
        """Monitor the Write Address channel for transactions
        """
        yield handshake(self.clock,
                        self.bus.AWREADY,
                        self.bus.AWVALID,
                        self.awready_delay)

        awaddr = int(self.bus.AWADDR)
        awlen = int(self.bus.AWLEN)
        awsize = int(self.bus.AWSIZE)
        awburst = int(self.bus.AWBURST)
        awid = int(self.bus.AWID)

        return (awaddr, awlen, awsize, awburst, awid)

    @cocotb.coroutine
    def _write_data(self):
        """Monitor the Write Data channel for transactions
        """
        last = 0
        wdata = []
        wstrb = []
        while not last:
            yield handshake(self.clock,
                            self.bus.WREADY,
                            self.bus.WVALID,
                            self.wready_delay)
            try:
                wdata.append(int(self.bus.WDATA))
            except ValueError:
                wdata.append(0)
            wstrb.append(int(self.bus.WSTRB))
            last = int(self.bus.WLAST)
            wid = int(self.bus.WID)  # should be the same for each beat

        return (wdata, wid, wstrb)

    @cocotb.coroutine
    def _write_transaction(self):
        """Process write transactions

        This requires both wdata and awaddr to complete.
        """
        while True:
            # Wait for wdata and awaddr
            data = cocotb.start_soon(self._write_data())
            addr = cocotb.start_soon(self._write_address())
            wdata, wid, wstrb = yield data.join()
            awaddr, awlen, awsize, awburst, awid = yield addr.join()

            self.log.debug(
                "Write Transaction:\n"
                "\t0x{:08X} <= {}"
                .format(
                    awaddr,
                    "\n\t{:14}".format('').join(
                        "0x{:016X}".format(d) for d in wdata
                    )
                )
            )

            cocotb.start_soon(self._write_response(
                awaddr, awlen, awsize, awburst, awid,
                wdata, wid, wstrb))

            # delay here until a thread is free
            while len(self._write_threads) >= self.max_threads:
                yield RisingEdge(self.clock)

    @cocotb.coroutine
    def _write_response(self, awaddr, awlen, awsize,
                        awburst, awid, wdata, wid, wstrb):
        """Wait for a write response
        """
        bresp = 0
        bresp_delay = self.bresp_delay
        try:
            assert awid == wid, "WID doesn't match AWID"
            assert awid not in self._write_threads, \
                "Thread ID already in progress"
            self._write_threads.append(awid)

            yield self._data_transfer(awaddr, 2**awsize, awlen + 1,
                                      awburst, wdata, wstrb)
        except Exception as e:
            self.log.warning("SAXI Error: {}".format(str(e)))
            bresp = 2
            bresp_delay = 1

        for i in range(bresp_delay - 1):
            yield RisingEdge(self.clock)

        yield self.bresp_lock.acquire()
        self.bus.BID.value = awid
        self.bus.BRESP.value = bresp
        yield handshake(self.clock, self.bus.BVALID, self.bus.BREADY)
        self.bresp_lock.release()

        if awid in self._write_threads:
            self._write_threads.remove(awid)

    @cocotb.coroutine
    def _read_transaction(self):
        """Process a read transaction
        """
        while True:
            yield handshake(self.clock,
                            self.bus.ARREADY,
                            self.bus.ARVALID,
                            self.arready_delay)

            araddr = int(self.bus.ARADDR)
            arlen = int(self.bus.ARLEN)
            arsize = int(self.bus.ARSIZE)
            arburst = int(self.bus.ARBURST)
            arid = int(self.bus.ARID)

            cocotb.start_soon(self._read_response(
                araddr, arlen, arsize, arburst, arid))

            # delay here until a thread is free
            while len(self._read_threads) >= self.max_threads:
                yield RisingEdge(self.clock)

    @cocotb.coroutine
    def _read_response(self, araddr, arlen, arsize, arburst, arid):
        """Wait for a read response, return the data transferred.
        """
        rresp = 0
        rresp_delay = self.rresp_delay
        try:
            assert arid not in self._read_threads, \
                "Thread ID already in progress"
            self._read_threads.append(arid)

            data = yield self._data_transfer(araddr, 2**arsize, arlen + 1,
                                             arburst)
        except Exception as e:
            self.log.warning("SAXI Error: {}".format(str(e)))
            rresp = 2
            rresp_delay = 1
            data = [0]

        for i in range(rresp_delay - 1):
            yield RisingEdge(self.clock)

        yield self.rresp_lock.acquire()
        for i, d in enumerate(data):
            self.bus.RID.value = arid
            self.bus.RRESP.value = rresp
            self.bus.RDATA.value = d
            self.bus.RLAST.value = (1 if i == len(data) - 1 else 0)
            yield handshake(self.clock, self.bus.RVALID, self.bus.RREADY)
            self.bus.RLAST.value = 0
        self.rresp_lock.release()

        if arid in self._read_threads:
            self._read_threads.remove(arid)

    @cocotb.coroutine
    def _data_transfer(self, start_address, num_bytes, burst_length, mode,
                       data=None, strbs=None):
        """Perform AXI4 Data Transfer

        This method is basically the pseudocode from A3.4.2 of the Protocol
        Specification
        """
        # [FIXED, INCR, WRAP]
        assert mode in [0, 1, 2], "Unsupported Burst Mode:0x{:1X}".format(mode)

        if data is None or strbs is None:
            is_write = False
            data = []
        else:
            is_write = True
            assert len(data) == burst_length, \
                "Data supplied doesn't match burst_length"
            assert len(strbs) == burst_length, \
                "Data supplied doesn't match burst_length"

        addr = start_address
        aligned_address = (addr // num_bytes) * num_bytes
        aligned = aligned_address == addr
        dtsize = num_bytes * burst_length
        lower_wrap_boundary = (addr // dtsize) * dtsize
        upper_wrap_boundary = lower_wrap_boundary + dtsize

        for i in range(burst_length):
            lower_byte_lane = addr - (addr // self._bus_width) \
                * self._bus_width
            if aligned:
                upper_byte_lane = lower_byte_lane + num_bytes - 1
            else:
                upper_byte_lane = aligned_address + num_bytes - 1 \
                    - (addr // self._bus_width) * self._bus_width

            # TODO check strb matches data address alignment
            if is_write:
                data_bytes = int.to_bytes(data[i], self._bus_width, 'little')
                yield self.do_slave_write(addr, data_bytes[lower_byte_lane:upper_byte_lane + 1])
            else:
                data_bytes = yield self.do_slave_read(addr, lower_byte_lane, upper_byte_lane)
                # TODO not sure how these bytes are aligned, might need some
                # padding
                data.append(int.from_bytes(data_bytes, 'little'))

            if mode != 0:
                if aligned:
                    addr = addr + num_bytes
                    if mode == 2:  # WRAP
                        if addr >= upper_wrap_boundary:
                            addr = lower_wrap_boundary
                else:  # INCR
                    addr = addr + num_bytes
                    aligned = True

        return data

    @cocotb.coroutine
    def do_slave_write(self, addr, data_bytes):
        """Perform action on a write transaction
        """
        return
        yield

    @cocotb.coroutine
    def do_slave_read(self, addr, low_byte, high_byte):
        """Perform action on a read transaction
        """
        return int.to_bytes(0, high_byte - low_byte + 1, 'little')
        yield


class AXI4MemoryMap(AXI4Slave):
    """An example AXI4 driver that provides a simple memory block.
    """
    def __init__(self, size, *args, **kwargs):
        AXI4Slave.__init__(self, *args, **kwargs)
        self._memory = memoryview(bytearray(size))

    @cocotb.coroutine
    def do_slave_write(self, addr, data_bytes):
        self._memory[addr:addr + len(data_bytes)] = data_bytes

        return
        yield

    @cocotb.coroutine
    def do_slave_read(self, addr, low_byte, high_byte):
        data_bytes = self._memory[addr:addr + high_byte + 1].tobytes()
        return data_bytes[low_byte:high_byte + 1]
        yield


class RegMap(AXI4LiteMaster):
    def writereg(self, reg, value):
        return self.write(reg << 2, value)

    def readreg(self, reg):
        return self.read(reg << 2)



import io
import re
import hashlib
import tarfile
import json
import logging
import asyncio
import concurrent.futures
from functools import cache
from pymoku.sdk import BuildStep
from pymoku.plugins import PluginCache
from pymoku.parts import SourceFile


_cache = PluginCache()
log = logging.getLogger("pymokusdk")


@cache
def get_platform(platform_name):
    """ Find a Platform checkpoint
    """
    manifests = _cache.checkpoint_manifests(_id=f'pymoku:{platform_name}')

    if not manifests:
        _cache.store_manifests(_cache.cloud_manifests(_type='checkpoint'))
    manifests = _cache.checkpoint_manifests(_id=f'pymoku:{platform_name}')

    if not manifests:
        log.error(f"No checkpoints found for {platform_name}")
        return None

    log.debug(f"Using platform {platform_name} from {manifests[0]['location']}")

    return CheckpointPlatform(platform_name, manifests)


class CheckpointPlatform(BuildStep):
    """ This mimics the vivado.Platform interface but loads the checkpoint from
        a barfile rather than trying to build it.

        This step downloads the barfile to the local pymoku cache; the
        CheckpointShell unpacks each DCP into the build directory.
    """
    def __init__(self, name, manifests):
        super().__init__(name=name)
        # there is no "platform" checkpoint, only the individual cells
        self.manifests = manifests

        self.hwver = self.slot_shells()[0].hwver
        self.part = self.slot_shells()[0].part
        self.arch = self.slot_shells()[0].arch
        # TODO these should not be platform properties
        self.slot_ains = self.slot_shells()[0].num_ain
        self.slot_aouts = self.slot_shells()[0].num_aout
        self.slot_cbufs = self.slot_shells()[0].num_cbufs
        self.sha = manifests[0]['plat_sha']

    @cache
    def slot_shells(self):
        shells = [CheckpointShell(self, m) for m in self.manifests]
        return sorted([s for s in shells if s.slot is not None], key=lambda s: s.slot)

    async def do_step(self):
        # manifest may be in one shared barfile, reduce to unique barfiles
        barfiles = []
        to_fetch = []
        for m in self.manifests:
            if m['location'] != 'local' and m['barfile'] not in barfiles:
                to_fetch.append(m)
                barfiles.append(m['barfile'])

        # split downloads across mutliple threads (and unblock other BuildSteps)
        tasks = []
        loop = asyncio.get_running_loop()
        with concurrent.futures.ThreadPoolExecutor(max_workers=5) as pool:
            for m in to_fetch:
                tasks.append(loop.run_in_executor(pool, _cache.download_barfile, m))
            await asyncio.gather(*tasks)


class CheckpointShell(BuildStep):
    def __init__(self, platform, manifest):
        self.platform = platform
        self.manifest = manifest
        self.hwver = manifest['extra']['hwver']
        self.cell_name = manifest['extra']['cell_name']
        self.pblock_name = manifest['extra']['pblock_name']
        self.part = manifest['extra']['part']
        self.arch = manifest['extra']['arch']
        self.num_ain = manifest['extra']['num_ain']
        self.num_aout = manifest['extra']['num_aout']
        self.num_cbufs = manifest['extra']['num_cbufs']

        name = f'{platform.name}-{self.safe_cell_name}'
        super().__init__(name=name)
        self.depends_on(platform)

        self.workdir = self.build_path / 'checkpoints' / platform.name
        self.shell_dcp = SourceFile(self.workdir / f'{name}.dcp')
        self.creates(self.shell_dcp)

    @property
    def plat_name(self):
        return self.platform.name

    @property
    def path(self):
        return next(iter(self.outputs())).path

    @property
    def safe_cell_name(self):
        name = self.cell_name.translate(
            str.maketrans({"[": "_", "]": None, "*": "_"}))
        name = name.replace('_INSTR', '')
        return name

    @property
    def slot(self):
        """ This makes some big assumptions about cell naming conventions"""
        if (m := re.search(r'INSTRS\[(\d+)\]', self.cell_name)) is None:
            return None
        return int(m.group(1))

    def config_state(self):
        sha = hashlib.sha256(json.dumps(self.manifest['extra']).encode())
        sha.update(self.manifest['sha256'].encode())
        sha.update(self.manifest['plat_sha'].encode())
        return sha.digest()

    async def do_step(self):
        barfile_data = _cache.download_barfile(self.manifest)
        barfile_data = io.BytesIO(barfile_data)

        with tarfile.open(fileobj=barfile_data) as arc:
            member = arc.getmember(self.manifest['filename'])
            assert member.isfile()
            data = arc.extractfile(member).read()

        sha = hashlib.sha256(data)

        assert sha.hexdigest() == self.manifest['sha256']

        self.shell_dcp.path.parent.mkdir(exist_ok=True, parents=True)
        self.shell_dcp.path.write_bytes(data)



import click
import logging
import tarfile
import json
import hashlib
import subprocess
from pathlib import Path
from rich.pretty import pprint
from rich.logging import RichHandler
from jinja2 import (Environment, PackageLoader, StrictUndefined,
                    select_autoescape)
from cryptography.exceptions import InvalidSignature
from cryptography.hazmat.primitives.asymmetric.ed25519 import Ed25519PublicKey
from base64 import b64decode
from pymoku.parts import include


FORMAT = "%(message)s"
logging.basicConfig(
    level="NOTSET", format=FORMAT, datefmt="[%X]", handlers=[RichHandler()]
)

log = logging.getLogger()


PLATFORMS = {  # TODO these could be parsed from part.defs
    'mokugo': [1, 2],
    'mokulab': [1, 2],
    'mokupro': [1, 4, 8],
    'mokudelta': [1, 4, 8],
}

templates = Path(__file__).parent / 'templates'
jenv = Environment(
    loader=PackageLoader('pymoku.sdk'),
    autoescape=select_autoescape(),
    undefined=StrictUndefined,
    keep_trailing_newline=True,
    trim_blocks=True,
)


def create_parts_def(path, name, namespace):
    path = path / 'parts.def'
    log.info(f"Create '{path}'")

    if path.exists():
        log.warning(f"'{path}' already exists, skipping")
        return

    tmpl = jenv.get_template('parts.def')
    text = tmpl.render(name=name.lower(),
                       namespace=namespace.lower(),
                       top_entity=f'{name}Slot')
    path.write_text(text)

    if (parts_path := Path('parts.def')).exists():
        include_text = f"include('{name.lower()}', prefix='{name.lower()}')"
        log.info(f'adding {include_text} to {parts_path}')
        with parts_path.open('a') as f:
            f.write(f"{include_text}\n")


def create_vhd_sources(path, name):
    top_entity = f'{name}Slot'
    path = path / f'{top_entity}.vhd'
    log.info(f"Create '{path}'")

    path.parent.mkdir(parents=True, exist_ok=True)

    if path.exists():
        log.warning(f"'{path}' already exists, skipping")
        return

    tmpl = jenv.get_template('Top.vhd')
    text = tmpl.render(top_entity=top_entity)
    path.write_text(text)


def append_platform(path, name, namespace, platform):
    path = path / 'parts.def'
    log.info(f"Adding {platform} to '{path}'")

    if not path.exists():
        log.error(f"'{path}' doesn't exist")
        return

    tmpl = jenv.get_template('parts_platform.def')
    text = tmpl.render(top_entity=f'{name}Slot',
                       name=name.lower(),
                       namespace=namespace.lower(),
                       hw=platform,
                       platforms=PLATFORMS[platform])

    with path.open('a') as f:
        f.write(text)


def create_slot_test(path, name):
    top_entity = f'{name}Slot'
    top_path = path / f'Test{top_entity}.vhd'
    log.info(f"Create '{top_path}'")

    top_path.parent.mkdir(parents=True, exist_ok=True)

    if not top_path.exists():
        tmpl = jenv.get_template('TestTop.vhd')
        text = tmpl.render(top_entity=top_entity, num_ain=4, num_aout=4)
        top_path.write_text(text)
    else:
        log.warning(f"'{top_path}' already exists, skipping")

    mod_path = path / f'test_{name.lower()}.py'
    log.info(f"Create '{mod_path}'")

    if not mod_path.exists():
        tmpl = jenv.get_template('test_module.py')
        text = tmpl.render(name=name.lower(), top_entity=top_entity, num_ain=4, num_aout=4)
        mod_path.write_text(text)
    else:
        log.warning(f"'{mod_path}' already exists, skipping")


def create_python_module(path, name):
    top_entity = f'{name}Slot'
    mod_path = path / f'{name.lower()}.py'
    log.info(f"Create '{mod_path}'")

    if not mod_path.exists():
        mod_path.parent.mkdir(parents=True, exist_ok=True)
        tmpl = jenv.get_template('module.py')
        text = tmpl.render(name=name, top_entity=top_entity)
        mod_path.write_text(text)
    else:
        log.warning(f"'{mod_path}' already exists, skipping")

    init_path = path / '__init__.py'
    log.info(f"Create '{init_path}'")

    if not init_path.exists():
        tmpl = jenv.get_template('plugin.py')
        text = tmpl.render(name=name, top_entity=top_entity)
        init_path.write_text(text)
    else:
        log.warning(f"'{init_path}' already exists, skipping")

    applet_path = path / 'applet.py'
    log.info(f"Create '{applet_path}'")

    if not applet_path.exists():
        tmpl = jenv.get_template('applet.py')
        text = tmpl.render(name=name)
        applet_path.write_text(text)
    else:
        log.warning(f"'{applet_path}' already exists, skipping")

    api_path = path / 'api.py'
    log.info(f"Create '{api_path}'")

    if not api_path.exists():
        tmpl = jenv.get_template('api.py')
        text = tmpl.render(name=name)
        api_path.write_text(text)
    else:
        log.warning(f"'{api_path}' already exists, skipping")


def init_git(path, name):
    gitignore_path = path / '.gitignore'
    log.info("Initializing Git repository")

    if not gitignore_path.exists():
        tmpl = jenv.get_template('gitignore')
        text = tmpl.render()
        gitignore_path.write_text(text)
    else:
        log.warning(f"'{gitignore_path}' already exists, skipping")

    subprocess.run(['git', 'init', '.'], cwd=path, capture_output=True)
    subprocess.run(['git', 'add', '-A'], cwd=path, capture_output=True)
    subprocess.run(['git', 'commit', '-m', f"Initial creation of {name}"], cwd=path, capture_output=True)


@click.group()
@click.version_option(package_name='pymoku.sdk')
def cli():
    pass


@cli.command()
@click.argument('name')
@click.option('--namespace', default='custom')
@click.option('--hw', help=f"Target hardware platform {list(PLATFORMS.keys())}")
@click.option('--git', is_flag=True, help="Initialize Git repository")
def new(name, namespace='custom', hw=None, git=False):
    path = Path('.') / name.lower()
    log.info(f"Creating new project '{path}'")
    path.mkdir(parents=True, exist_ok=True)

    create_parts_def(path, name, namespace)
    create_vhd_sources(path / 'vhdl', name)
    create_python_module(path / 'python', name)
    create_slot_test(path / 'tests', name)

    if hw:
        append_platform(path, name, namespace, hw)

    if git:
        init_git(path, name)

    log.info('Finished')


@cli.command()
@click.argument('name')
@click.argument('hw')
def add_platform(name, hw):
    path = Path('.') / name.lower()
    append_platform(path, name, hw)


@cli.command()
@click.argument('name')
def build(name):
    build_targets = include(Path(name) / 'parts.def')
    print(build_targets)


from pymoku.sdk import keys  # noqa


@cli.command()
@click.argument('barfile', type=click.Path(exists=True, dir_okay=False))
@click.option('--raw', is_flag=True, help="Don't decode or pretty-print")
def manifest(barfile, raw=False):
    """ Print the manifest from the specified barfile. """
    with tarfile.open(barfile, mode='r') as tf:
        manifest = tf.extractfile('MANIFEST')
        if raw:
            print(manifest.read().decode())
        else:
            pprint(json.load(manifest))


@cli.command()
@click.argument('barfile', type=click.Path(exists=True, dir_okay=False))
def verify(barfile):
    """ Verify all manifest signatures available and all file hashes of the
        specified barfile.
   """
    with tarfile.open(barfile, mode='r') as tf:
        manifest_bytes = tf.extractfile('MANIFEST').read()
        manifest = json.loads(manifest_bytes.decode())

        # check all the signatures match
        for keyname, pubkey_b64 in keys.list_public_keys():
            pubkey = Ed25519PublicKey.from_public_bytes(b64decode(pubkey_b64))
            try:
                sig_bytes = tf.extractfile(f'MANIFEST.{keyname}.sig').read()
                pubkey.verify(sig_bytes, manifest_bytes)
                log.info(f"Signature matches '{keyname}'")
            except KeyError:
                log.debug(f"no signature for '{keyname}'")
            except InvalidSignature:
                log.error(f"Invalid signature for '{keyname}'")

        # check all the file hashes
        for item in manifest['items']:
            if (fname := item.get('filename')) is None:
                continue

            if (manifest_sha := item.get('sha256')) is None:
                continue

            log.debug(f"Checking hash for '{item.get('_id')}:{fname}'")
            try:
                file_sha = hashlib.sha256(tf.extractfile(fname).read())
                if file_sha.hexdigest() != manifest_sha:
                    log.error(f"Hash does not match for '{item.get('_id')}'")
            except KeyError:
                log.error(f"Referenced file '{item.get('_id')}:{fname}' not in archive")


if __name__ == '__main__':
    cli()



import itertools
import tarfile
import json
import io
import hashlib
from pathlib import Path
from pymoku.parts import BuildStep, SourceFile
from pymoku.sdk import SCM_VERSION, log, keys


KEY_DIR = Path.home() / '.config/pymoku/keys'


class BarFile(BuildStep):
    VERSION = 0

    def __init__(self, barfile, *manifests):
        super().__init__(name=barfile)

        # list of manifests or list of lists of manifests
        try:
            self.manifests = list(itertools.chain.from_iterable(manifests))
        except TypeError:
            self.manifests = manifests

        self.barfile = self.build_path / barfile
        for m in self.manifests:
            self.depends_on(m.bitstreams())
        self.creates(SourceFile(self.barfile))
        self._bs_arcnames = {}

    def arcname_for(self, bitstream):
        # generate a unique filename for bitstreams in the archive
        if bitstream not in self._bs_arcnames:
            self._bs_arcnames[bitstream] = f'{len(self._bs_arcnames):03d}'
        return self._bs_arcnames[bitstream]

    def config_state(self):
        # Manifests aren't build steps, and we don't know when they change
        # until their own dependencies are built.
        sha = hashlib.sha256(SCM_VERSION.encode())
        sha.update(json.dumps(dict(keys.list_public_keys())).encode())
        return sha.digest()

    async def do_step(self):
        manifests = []
        for m in self.manifests:
            manifests.extend(m(self.arcname_for))

        self.update_progress(total=2 + len(self._bs_arcnames))

        manifest_bytes = self.create_manifest([m for m, bs in manifests])
        self.update_progress(advance=1)

        with tarfile.open(self.barfile, mode='w') as tf:
            m_tinfo = tarfile.TarInfo(name='MANIFEST')
            m_tinfo.size = len(manifest_bytes)
            tf.addfile(m_tinfo, io.BytesIO(manifest_bytes))

            # sign with all the keys TODO maybe specify which ones?
            for keyname, private_key in keys.list_private_keys():
                sig_bytes = private_key.sign(manifest_bytes)
                sig_tinfo = tarfile.TarInfo(name=f'MANIFEST.{keyname}.sig')
                sig_tinfo.size = len(sig_bytes)
                tf.addfile(sig_tinfo, io.BytesIO(sig_bytes))

            for bs, fname in self._bs_arcnames.items():
                bs_info = tf.gettarinfo(bs.path, arcname=fname)
                with open(bs.path, 'rb') as f:
                    tf.addfile(bs_info, fileobj=f)
                self.update_progress(advance=1)

        # 32 MB is the upload limit through the web interface, but also probably
        # a reasonable limit in general
        size = self.barfile.stat().st_size
        if size > 1024 * 2**15:
            log.warning(f'Barfile {self.barfile} exceeds 32MB ({size:d})')

    def create_manifest(self, items):
        manifest = dict(version=self.VERSION, items=items)
        return json.dumps(manifest, indent=2).encode()



import struct
import cocotb
from cocotb.clock import Clock
from cocotb.triggers import RisingEdge, FallingEdge, Timer
from cocotb_bus.drivers import BusDriver


class SPIRegMap(BusDriver):
    _signals = ["SPI_MISO", "SPI_MOSI", "SPI_CS", "SPI_SClk"]

    def __init__(self, instr, name=None, freq=1e6, **kwargs):
        BusDriver.__init__(self, instr, name, instr.clk, **kwargs)
        self.freq = round(1e9 / freq, 4)

        # Drive some sensible defaults (setimmediatevalue to avoid x asserts)
        self.bus.SPI_MOSI.setimmediatevalue(0)
        self.bus.SPI_CS.setimmediatevalue(1)
        self.bus.SPI_SClk.setimmediatevalue(0)

    async def writereg(self, reg, value):
        b = struct.pack('<II', (reg << 2) | 0x80000000, value)
        res = await self.spi_xfer2(b)

    async def writeregs(self, regs):
        for r, v, in regs:
            b = struct.pack('<II', (r << 2) | 0x80000000, v)
            await self.spi_xfer2(b)

    async def readreg(self, reg):
        b = struct.pack('<II', reg << 2, 0)
        res = await self.spi_xfer2(b)
        return struct.unpack('<II', bytes(res))[1]

    async def writeblock(self, addr, data):
        b = struct.pack('<I', addr | 0x80000000) + data
        res = await self.spi_xfer2(b)

    async def readblock(self, addr, length):
        b = struct.pack('<I', addr) + bytes([0] * length)
        res = bytes(await self.spi_xfer2(b))
        return res[4:]

    async def spi_xfer2(self, data):
        """ should look like SpiDev.xfer2()
        """
        self.bus.SPI_CS.value = 0
        self.bus.SPI_MOSI.value = (data[0] & 0x80) >> 7

        # this timing changes wildly between spi controllers
        await Timer(time=0.3, units='us', round_mode='round')
        sclk_task = cocotb.start_soon(
            Clock(self.bus.SPI_SClk, self.freq, units='ns').start(cycles=len(data) * 8))

        miso_data = []

        for d in data:
            m = 0
            for i in range(8):
                self.bus.SPI_MOSI.value = (d & 0x80) >> 7
                d = d << 1
                await RisingEdge(self.bus.SPI_SClk)
                m = (m << 1) | self.bus.SPI_MISO.value
                await FallingEdge(self.bus.SPI_SClk)
            miso_data.append(m)

        await sclk_task
        await Timer(time=0.3, units='us', round_mode='round')
        self.bus.SPI_CS.value = 1
        await Timer(time=0.3, units='us', round_mode='round')
        return miso_data



import os
import struct
import pymoku
import pymoku.slot
from pymoku.network.render import CacheableRenderControl
from pymoku.registers import RegisterAccess
from spiregs import SPIRegMap


class SimRenderControl(CacheableRenderControl):
    def __init__(self):
        self._refs = []

        self._cached_state = {
            'cbuf0': {
                '_phy_offset': 34359738368,
                '_phy_size': 6442450944,
            },
        }
        self._cached_write = {}
        self._cached = True


class CoCoTBSlot(pymoku.Slot):
    """ Abuse the cacheable layer to gather write commands
        call await apply() from cocotb context to write regs
    """
    def __init__(self, dut):
        RegisterAccess.__init__(self)
        self._Slot__inputs = {}
        self._Slot__outputs = {}
        self.dut = dut
        self.spi = None
        self._CachableAccessControl__cached = True
        self.num_regs = 128
        self.slot = 0
        block = b'\x00' * self.num_regs * 4
        self._CachableAccessControl__cached_regs = list(struct.unpack(f'<{self.num_regs:d}I', block))
        self._CachableAccessControl__to_write = []
        self.rdr_ctrl = SimRenderControl()
        self.init_regs(**self.parse_generics())

    def parse_generics(self):
        generics = os.getenv('TEST_GENERICS').split(' ')
        generics = [g.split('=') for g in generics]
        g = dict()
        for k, v in generics:
            # try to guess the type, we can only do so much...
            k = k.lower()
            try:
                g[k] = int(v)  # integer
            except ValueError:
                if v.lower() in ['true', 'false']:
                    g[k] = str(v.lower() == 'true')  # boolean
                else:
                    g[k] = v  # string

        self.num_ain = g['num_ain']
        self.num_aout = g['num_aout']
        self.num_cbufs = g['num_cbufs']
        return g

    def init_regs(self, **kwargs):
        pass

    def cbufs(self):
        return [f'cbuf{i:d}' for i in range(self.num_cbufs)]

    async def apply(self):
        if self.spi is None:
            self.spi = SPIRegMap(self.dut, freq=50e6)

        if self._CachableAccessControl__to_write:
            await self.spi.writeregs(self._CachableAccessControl__to_write)

        # block = await self.spi.readblock(0, 4 * self.num_regs)
        # self._CachableAccessControl__cached_regs = list(struct.unpack(f'<{self.num_regs:d}I', block))
        self._CachableAccessControl__to_write = []

    def _fetch_state(self):
        pass

    def _close_state(self):
        pass



# monkey patch
pymoku.Slot = CoCoTBSlot
pymoku.slot.Slot = CoCoTBSlot


# import the actual testbench
star = {k: v for k, v in vars(__import__(os.getenv('TEST_MODULE'), fromlist=['*'])).items()
        if not k.startswith('_')}
globals().update(star)


class SimMoku(object):
    num_slots = 1
    AXI_BUS_WIDTH = 16

    def sys_ref(self):
        return CLK_FREQ  # from the cocotb testbench

    def adc_sample_rate(self):
        return self.sys_ref() * 4.0


CoCoTBSlot.moku = SimMoku()



import sys
import logging
import zmq
import pymoku
# jsonrpcserver<5 included in firmware
import jsonrpcserver as rpc


__version__ = '1.0'
INSTR_ID = sys.argv[1]
SLOT = int(sys.argv[2])

logging.basicConfig(level=logging.DEBUG)
log = logging.getLogger(f'slot{SLOT:d}-rpc')


@rpc.method
def get_info(instr):
    data = dict(version=__version__,
                instrid=INSTR_ID,
                slot=instr.slot,
                hw=type(instr.moku).__name__,
                serial=instr.moku.serial)
    return data


@rpc.method
def dump_regs(instr):
    return instr.dump_regs()


def serve(instr, addr):
    skt = zmq.Context().socket(zmq.REP)
    skt.bind(addr)

    while True:
        req = skt.recv_string()
        log.debug(req)
        skt.send_string(str(rpc.dispatch(req, context=instr)))


def main():
    log.info(f'starting {INSTR_ID}:{SLOT:d}')

    # We can't rely on pymoku's plugin loading, as it needs python>=3.10
    # but we're already 'in' the plugin, se we can load things manually.
    moku = pymoku.connect_ipc(attach=False)

    # constructing the instrument directly assumes it has been deployed already
    from __pymoku_plugin import Plugin
    m = dict(slot=SLOT)  # TODO one day the manifest may do more here
    slot = Plugin().attach(moku, manifest=m)

    from __pymoku_plugin import api

    addr = f'ipc://./instr{SLOT:d}'
    log.info(f'serving {type(slot).__name__} @ {addr}')
    serve(slot, addr)


main()



import struct
import cocotb
import queue
from cocotb.triggers import RisingEdge


def find_handle(root, path):
    """ Find a hierachy node given a list of node names
        super hacky around cocotb's broken generate/block support
        ampl_lut = find_handle(dut.dut, ['chans(0)', 'chan', 'netan_gen', 'ampl_lut'])
    """
    root._discover_all()
    if not path:
        return root
    return find_handle(root._sub_handles[path.pop(0)], path)


async def _fifo_write(data, clk, en, din, full, overflow=None,
                      prog_full=None, prog_full_threshold=0):
    while True:
        await RisingEdge(clk)
        if en.value:
            data.put_nowait(din.value)

        full.value = int(data.full())  # TODO the delay thing
        if overflow is not None:
            if en.value.is_resolvable:
                overflow.value = int(en.value and data.full())

        if prog_full is not None:
            prog_full.value = int(data.qsize() > prog_full_threshold)


async def _fifo_read(data, clk, en, dout, empty, valid=None, length=1024,
                     prog_empty=None, prog_empty_threshold=0):
    rdq = queue.Queue(length)

    async def _dly(val):
        await RisingEdge(clk)
        await RisingEdge(clk)
        rdq.put_nowait(val)

    while True:
        await RisingEdge(clk)
        empty.value = int(rdq.empty())

        if prog_empty is not None:
            prog_empty.value = int(rdq.qsize() < prog_empty_threshold)

        if not data.empty() and not rdq.full():
            cocotb.start_soon(_dly(data.get_nowait()))

        if en.value:
            if rdq.empty():
                dout.binstr = 'zzzzzzzz'
            else:
                dout.value = rdq.get_nowait()
                if valid is not None:
                    valid.value = 1
        else:
            if valid is not None:
                valid.value = 0


def sim_fifo(dut, length=1024):
    # rst : in std_logic;
    # wr_clk : in std_logic;
    # rd_clk : in std_logic;
    # din : in std_logic_vector(47 downto 0);
    # wr_en : in std_logic;
    # rd_en : in std_logic;
    # dout : out std_logic_vector(11 downto 0);
    # full : out std_logic;
    # empty : out std_logic;
    # valid : out std_logic
    data = queue.Queue(length)
    cocotb.start_soon(_fifo_write(data, dut.wr_clk, dut.wr_en, dut.din, dut.full))
    cocotb.start_soon(_fifo_read(data, dut.rd_clk, dut.rd_en, dut.dout, dut.empty, length=length))



def cbuf_fifo(handle, length=1024):
    # rst : in std_logic;
    # wr_clk : in std_logic;
    # rd_clk : in std_logic;
    # din : in std_logic_vector(107 downto 0);
    # wr_en : in std_logic;
    # rd_en : in std_logic;
    # dout : out std_logic_vector(107 downto 0);
    # full : out std_logic;
    # overflow : out std_logic;
    # empty : out std_logic;
    # prog_full : out std_logic;
    # prog_empty : out std_logic;
    # wr_rst_busy : out std_logic;
    # rd_rst_busy : out std_logic
    data = queue.Queue(length)  # TODO reset
    cocotb.start_soon(_fifo_write(data, handle.wr_clk, handle.wr_en, handle.din, handle.full,
                      handle.overflow, handle.prog_full, prog_full_threshold=length // 2))
    cocotb.start_soon(_fifo_read(data, handle.rd_clk, handle.rd_en, handle.dout, handle.empty, length=length,
                      prog_empty=handle.prog_empty, prog_empty_threshold=16))
    handle.wr_rst_busy.value = 0
    handle.rd_rst_busy.value = 0


def sim_fft(dut):
    # aclk : in std_logic;
    # aresetn : in std_logic;
    # s_axis_config_tdata : in std_logic_vector(7 downto 0);
    # s_axis_config_tvalid : in std_logic;
    # s_axis_config_tready : out std_logic;
    # s_axis_data_tdata : in std_logic_vector(31 downto 0);
    # s_axis_data_tvalid : in std_logic;
    # s_axis_data_tready : out std_logic;
    # s_axis_data_tlast : in std_logic;
    # m_axis_data_tdata : out std_logic_vector(63 downto 0);
    # m_axis_data_tuser : out std_logic_vector(15 downto 0);
    # m_axis_data_tvalid : out std_logic;
    # m_axis_data_tready : in std_logic;
    # m_axis_data_tlast : out std_logic;
    # event_frame_started : out std_logic;
    # event_tlast_unexpected : out std_logic;
    # event_tlast_missing : out std_logic;
    # event_status_channel_halt : out std_logic;
    # event_data_in_channel_halt : out std_logic;
    # event_data_out_channel_halt : out std_logic

    dut.s_axis_config_tready.value = 0
    dut.s_axis_data_tready.value = 0
    dut.m_axis_data_tdata.value = 0
    dut.m_axis_data_tuser.value = 0
    dut.m_axis_data_tvalid.value = 0
    dut.m_axis_data_tlast.value = 0
    dut.event_frame_started.value = 0
    dut.event_tlast_unexpected.value = 0
    dut.event_tlast_missing.value = 0
    dut.event_status_channel_halt.value = 0
    dut.event_data_in_channel_halt.value = 0
    dut.event_data_out_channel_halt.value = 0


async def _ram(data, clk, addr, dout=None, en=None, we=None, din=None):
    async def _dly(a, val):
        await RisingEdge(clk)
        data[a] = val

    while True:
        try:
            await RisingEdge(clk)
            if en is None or en.value:
                a = addr.value
                if dout:
                    dout.value = data[a]
                if we is not None and we.value:
                    cocotb.start_soon(_dly(a, din.value))
        except ValueError:
            pass


def arbwave_buffer(handle):
    # clka : in std_logic;
    # ena : in std_logic;
    # wea : in std_logic_vector(0 to 0);
    # addra : in std_logic_vector(12 downto 0);
    # dina : in std_logic_vector(17 downto 0);
    # douta : out std_logic_vector(17 downto 0);
    # clkb : in std_logic;
    # enb : in std_logic;
    # web : in std_logic_vector(0 to 0);
    # addrb : in std_logic_vector(12 downto 0);
    # dinb : in std_logic_vector(17 downto 0);
    # doutb : out std_logic_vector(17 downto 0)

    data = [0] * 0x1000
    cocotb.start_soon(_ram(data, handle.clka, handle.addra, handle.douta, handle.ena, handle.wea, handle.dina, "READ_FIRST",
                                 handle.clkb, handle.addrb, handle.doutb, handle.enb, handle.web, handle.dinb, "READ_FIRST"))


def sin_lut(handle):
    # clka : in std_logic;
    # clkb : in std_logic;
    # addra : in std_logic_vector(9 downto 0);
    # douta : out std_logic_vector(17 downto 0);
    # addrb : in std_logic_vector(9 downto 0);
    # doutb : out std_logic_vector(17 downto 0)
    import sin_lut
    data = sin_lut.gen_lut(length=2**10, bitdepth=18)
    cocotb.start_soon(_ram(data, handle.clka, handle.addra, handle.douta, handle.ena, None, None, "WRITE_FIRST",
                                 handle.clkb, handle.addrb, handle.doutb, handle.enb, None, None, "WRITE_FIRST"))


def window_lut(handle):
    # clka : in std_logic;
    # ena : in std_logic;
    # wea : in std_logic_vector(0 to 0);
    # addra : in std_logic_vector(12 downto 0);
    # dina : in std_logic_vector(17 downto 0);
    # douta : out std_logic_vector(17 downto 0);
    # clkb : in std_logic;
    # enb : in std_logic;
    # web : in std_logic_vector(0 to 0);
    # addrb : in std_logic_vector(12 downto 0);
    # dinb : in std_logic_vector(17 downto 0);
    # doutb : out std_logic_vector(17 downto 0)
    import window_functions
    data = window_functions.bhwindow()
    cocotb.start_soon(_ram(data, handle.clka, handle.addra, handle.douta, handle.ena, handle.wea, handle.dina, "WRITE_FIRST",
                                 handle.clkb, handle.addrb, handle.doutb, handle.enb, handle.web, handle.dinb, "READ_FIRST"))


def accum_buffer(handle):
    # clka : in std_logic;
    # wea : in std_logic_vector(0 downto 0);
    # addra : in std_logic_vector(11 downto 0);
    # dina : in std_logic_vector(53 downto 0);
    # clkb : in std_logic;
    # addrb : in std_logic_vector(11 downto 0);
    # doutb : out std_logic_vector(53 downto 0)

    data = [0] * 0x1000
    cocotb.start_soon(_ram(data,
        handle.clka, handle.addra, None, None, handle.wea, handle.dina, "READ_FIRST",
        handle.clkb, handle.addrb, handle.doutb, None, None, None, "READ_FIRST"))



try:
    import setuptools_scm
    SCM_VERSION = setuptools_scm.get_version()
except ImportError:
    SCM_VERSION = ''

FIRMWARE_VERSION = 616


from .vivado import *  # noqa
from .manifests import *  # noqa
from .barfile import *  # noqa
from .cocotb import *  # noqa
from .checkpoints import *  # noqa


def generate_slot_bitstreams(platform, synth,
                             constraints='support-slot-constraints',
                             route_tcl='support-slot-route-tcl',
                             bitstream_tcl='support-slot-bitstream-tcl',
                             reports_tcl='support-slot-reports-tcl'):
    """
    There is a distinction between arbitrary RMs and Slot RMs, in that Slot RMs are
    expected to be implemented within the group of Slots defined as "Instruments",
    whereas any RM can be implemented within any Slot as long as it meets the
    interface.

    This API simply provides a shortcut to the common case of "Instruments".
    """

    routes = [RMRoute(synth, shell, route_tcl=route_tcl, constraints=constraints) for shell in platform.slot_shells()]
    rms = [RMInstance(route, bitstream_tcl=bitstream_tcl, reports_tcl=reports_tcl) for route in routes]
    return rms


def generate_slot_manifests(platform, synth, manifest_id,
                            constraints='support-slot-constraints',
                            route_tcl='support-slot-route-tcl',
                            bitstream_tcl='support-slot-bitstream-tcl',
                            reports_tcl='support-slot-reports-tcl'):
    rms = generate_slot_bitstreams(platform, synth,
                                   constraints=constraints,
                                   route_tcl=route_tcl,
                                   bitstream_tcl=bitstream_tcl,
                                   reports_tcl=reports_tcl)
    manifests = [Manifest(manifest_id, rm) for rm in rms]
    return manifests


def include_support():
    import pymoku.sdk  # noqa
    import shutil  # noqa
    from pymoku.parts import include, ConfigGlobals  # noqa
    from importlib import resources  # noqa

    support_path = ConfigGlobals.build_path / 'support'
    if not support_path.exists():
        with resources.path(pymoku.sdk, 'support') as support_res:
            print(support_res, support_path)
            shutil.copytree(support_res, support_path)
    include(support_path, prefix='support')



import click
import logging
import json
from pathlib import Path
from base64 import b64encode
from cryptography.hazmat.primitives.asymmetric.ed25519 import Ed25519PrivateKey
from rich.console import Console
from rich.table import Table
from rich.pretty import pprint
from .main import cli


log = logging.getLogger("pymokusdk")


KEY_DIR = Path.home() / '.config/pymoku/keys'


def generate_private_key(name, overwrite=False):
    private_key = Ed25519PrivateKey.generate()
    KEY_DIR.mkdir(parents=True, exist_ok=True)
    keyfile = KEY_DIR / f'{name}.key'

    if keyfile.exists() and not overwrite:
        raise Exception(f"private key '{name}' already exists")

    keyfile.write_bytes(private_key.private_bytes_raw())
    log.info(f'wrote private key to {keyfile}')

    public_key = private_key.public_key()
    public_key_64 = b64encode(public_key.public_bytes_raw()).decode()
    return public_key_64


def load_private_key(keyname):
    key_bytes = (KEY_DIR / keyname).with_suffix('.key').read_bytes()
    return Ed25519PrivateKey.from_private_bytes(key_bytes)


def list_private_keys():
    for keyfile in KEY_DIR.glob('*.key'):
        private_key = load_private_key(keyfile.stem)
        yield (keyfile.stem, private_key)


def list_public_keys():
    for keyfile in KEY_DIR.glob('*.key'):
        private_key = load_private_key(keyfile.stem)
        public_key = private_key.public_key()
        public_key_64 = b64encode(public_key.public_bytes_raw()).decode()
        yield (keyfile.stem, public_key_64)


@cli.group()
def keys():
    pass


@keys.command('create')
@click.argument('name')
@click.option('-f', '--force', is_flag=True)
def create_key(name, force):
    public_key = generate_private_key(name, overwrite=force)
    pprint(f'Public Key: {name} {public_key}')


@keys.command('delete')
@click.argument('name')
def delete_key(name):
    keyfile = KEY_DIR / f'{name}.key'
    keyfile.unlink(missing_ok=True)


@keys.command('list')
@click.option('--json', '_json', is_flag=True, help='output keys as JSON')
def list_keys(_json=False):
    if _json:
        print(json.dumps(dict(list_public_keys())))
    else:
        table = Table()

        table.add_column("Key Name", justify="right", no_wrap=True)
        table.add_column("Public Key (base64)", no_wrap=True)

        for name, pubkey in list_public_keys():
            table.add_row(name, pubkey)

        console = Console()
        console.print(table)



import re
import hashlib
import zipfile
from pathlib import Path
from functools import partial, cache
from dataclasses import dataclass
from pymoku.parts import (
    log, link, copy, BuildStep, ProcessStep, SourceFile, ParseNode,
    global_alias, arg)


show_stats = arg('--stats', action='store_true')


_with_rich = True
try:
    from rich.table import Table
    from rich.bar import Bar
    from rich.text import Text
    from rich.columns import Columns
except ImportError:
    _with_rich = False


class VivadoStep(ProcessStep):
    def __init__(self, script, *args, **kwargs):
        self.dcp = kwargs.pop('dcp', None)
        self.tcl_args = kwargs.pop('tcl_args', ())
        super().__init__(**kwargs)
        # resolve script to an actual path
        self.script = list(self.depends_on(link(script, self.workdir)))[0].path

    def parse_tree(self):
        fn = self.update_progress
        start_node = ParseNode()
        node = start_node
        node = node.add(r'source.*\.tcl', partial(fn, completed=2, phase='running'))
        node = node.add(r'INFO:.*Exiting Vivado', partial(fn, completed=99))
        return start_node

    def process_log(self, logline):
        line = logline.decode('utf8').strip()
        if re.match(r'WARNING', line):
            self.log.warning(line)
        elif re.match(r'CRITICAL WARNING', line):
            self.log.error(line)
        elif re.match(r'ERROR', line):
            self.log.error(line)
        return super().process_log(logline)

    def _arg_strings(self):
        """ Convert args to string to pass to subprocess
        """
        for a in self.tcl_args:
            if isinstance(a, SourceFile):
                a = self.relative_to(self.config_path / a.path)  # TODO figure out this config_path
            yield str(a)

    def command(self):
        cmd = f'vivado -mode batch -source {self.relative_to(self.script)}'
        cmd += f' {self.dcp}' if self.dcp else ''
        if self.tcl_args:
            cmd += f' -tclargs {" ".join(self._arg_strings())}'
        return cmd


class ProcessorStep(VivadoStep):
    def parse_tree(self):
        example_log = [
            (r'source.*\.tcl', dict(completed=2)),
            (r'.*create_root_design', dict(completed=12, phase='running')),
            (r'VHDL Output written to', dict(completed=13)),
            (r'Starting synth_design', dict(completed=21)),
            (r'Starting RTL Elaboration', dict(completed=25)),
            (r'Start Handling Custom Attributes', dict(completed=32)),
            (r'Processing XDC Constraints', dict(completed=36)),
            (r'Completed Processing XDC Constraints', dict(completed=39)),
            (r'Finished Constraint Validation', dict(completed=41)),
            (r'Finished RTL Optimization Phase 2', dict(completed=50)),
            (r'Start RTL Component Statistics', dict(completed=52)),
            (r'Start Applying XDC Timing Constraints', dict(completed=65)),
            (r'Start Timing Optimization', dict(completed=66)),
            (r'Start Technology Mapping', dict(completed=73)),
            (r'Start IO Insertion', dict(completed=81)),
            (r'Start Final Netlist Cleanup', dict(completed=82)),
            (r'Finished Final Netlist Cleanup', dict(completed=84)),
            (r'Start Renaming Generated Instances', dict(completed=85)),
            (r'Start Rebuilding User Hierarchy', dict(completed=86)),
            (r'Start Renaming Generated Ports', dict(completed=87)),
            (r'synth_design completed successfully', dict(completed=95)),
            (r'INFO:.*Exiting Vivado', dict(completed=99)),
        ]

        start_node = ParseNode()
        fn = self.update_progress
        node = start_node
        for msg, kwargs in example_log:
            node = node.add(msg, partial(fn, **kwargs))

        return start_node


class SynthStep(VivadoStep):
    def parse_tree(self):
        example_log = [
            (r'source.*\.tcl', dict(completed=1)),
            (r'Starting synth_design', dict(completed=3, phase='running')),
            (r'Starting RTL Elaboration', dict(completed=4)),
            (r'Start Handling Custom Attributes', dict(completed=6)),
            (r'Netlist sorting complete', dict(completed=7)),
            (r'Processing XDC Constraints', dict(completed=10)),
            (r'Completed Processing XDC Constraints', dict(completed=12)),
            (r'Finished Constraint Validation', dict(completed=14)),
            (r'Finished RTL Optimization Phase 2', dict(completed=30)),
            (r'Start RTL Component Statistics', dict(completed=33)),
            (r'Finished RTL Component Statistics', dict(completed=33)),
            (r'Start Part Resource Summary', dict(completed=33)),
            (r'Finished Part Resource Summary', dict(completed=33)),
            (r'Start Cross Boundary and Area Optimization', dict(completed=33)),
            (r'Finished Cross Boundary and Area Optimization', dict(completed=47)),
            (r'Start Applying XDC Timing Constraints', dict(completed=48)),
            (r'Start Timing Optimization', dict(completed=50)),
            (r'Start Technology Mapping', dict(completed=58)),
            (r'Start IO Insertion', dict(completed=70)),
            (r'Start Final Netlist Cleanup', dict(completed=70)),
            (r'Finished Final Netlist Cleanup', dict(completed=73)),
            (r'Finished IO Insertion', dict(completed=74)),
            (r'Start Renaming Generated Instances', dict(completed=74)),
            (r'Start Handling Custom Attributes', dict(completed=78)),
            (r'Start Writing Synthesis Report', dict(completed=78)),
            (r'Synthesis Optimization Runtime', dict(completed=79)),
            (r'Netlist sorting complete', dict(completed=83)),
            (r'Reading XDEF placement', dict(completed=92)),
            (r'synth_design completed successfully', dict(completed=93)),
            (r'Writing XDEF routing', dict(completed=95)),
            (r'INFO:.*Exiting Vivado', dict(completed=99)),
        ]

        start_node = ParseNode()
        fn = self.update_progress
        node = start_node
        for msg, kwargs in example_log:
            node = node.add(msg, partial(fn, **kwargs))

        return start_node


class RouteStep(VivadoStep):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.wns = 0.0
        self.tns = 0.0
        self.phase = ''
        self.overlaps = ''

    def stats(self, match):
        self.wns = float(match.group('wns'))
        self.tns = float(match.group('tns'))
        log.debug(f'{self} WNS={self.wns:6.3f} TNS={self.tns:6.3f}')
        col = '[red]' if self.tns < 0.0 else ''
        self.update_progress(comment=f'{col}WNS={self.wns:6.3f} TNS={self.tns:6.3f}')

    def match_phase(self, match):
        self.phase = match.group('phase')
        log.debug(f'{self} Phase: {self.phase}')

    def match_overlaps(self, match):
        self.overlaps = match.group('overlaps')
        log.debug(f'{self} Overlaps: {self.overlaps}')
        self.update_progress(comment=f'WNS={self.wns:6.3f} TNS={self.tns:6.3f} {self.overlaps}')

    def process_log(self, logline):
        line = logline.decode('utf8')

        if (match := re.match(r'.*WNS=(?P<wns>[-\d]+\.[\d]+)[\s\|]*TNS=(?P<tns>[-\d]+\.[\d]+)', line)) is not None:
            self.stats(match)

        if (match := re.match(r'Phase (?P<phase>[\d]+[\.\d]*)', line)) is not None:
            self.match_phase(match)

        if (match := re.match(r'\sNumber of Nodes with overlaps = (?P<overlaps>[\d]+)', line)) is not None:
            self.match_overlaps(match)

        return super().process_log(logline)

    def parse_tree(self):
        example_log = [
            (r'Command: open_checkpoint', dict(completed=0, phase='open_checkpoint')),
            (r'Command: opt_design', dict(completed=3, phase='opt_design')),
            (r'Phase 1 Initialization', dict(completed=4)),
            (r'Phase 1.2 Setup Constraints And Sort Netlist', dict(completed=9)),
            (r'Phase 6 BUFG optimization', dict(completed=10)),
            (r'Command: place_design', dict(completed=10, phase='place_design')),
            (r'Command: phys_opt_design', dict(completed=30, phase='phys_opt_design')),
            (r'Command: route_design', dict(completed=50, phase='route_design')),
            (r'Command: phys_opt_design', dict(completed=90, phase='phys_opt_design')),
        ]

        start_node = ParseNode(strict=True)
        fn = self.update_progress
        node = start_node
        for msg, kwargs in example_log:
            node = node.add(msg, partial(fn, **kwargs))

        return start_node


# class GenIPStep(VivadoStep):
#     def __init__(self, xci):
#         self.xci = xci
#         name = f'{xci.path.stem}'
#         super().__init__('scripts/ip.tcl', xci.path.name, name=name)
#         self.depends_on(Copy(xci, XCIFile(self.workdir / xci.path.name)))

#     # def workdir(self):
#     #     return self.xci.outputs()[0].path.parent

#     def outputs(self):
#         return [DCPFile(self.workdir / self.xci.path.with_suffix('.dcp'))]

#     def parse_tree(self):
#         fn = self.update_progress
#         start_node = ParseNode()
#         node = start_node.add(r'source.*ip\.tcl', partial(fn, completed=5, phase='starting'))
#         node = node.add(r'Starting synth_design', partial(fn, completed=10, phase='synth'))
#         node = node.add(r'Starting RTL Elaboration', partial(fn, completed=30, phase='elaborate'))
#         node = node.add(r'Finished RTL Elaboration', partial(fn, completed=60, phase='finishing'))
#         return start_node


# @expects('support-processor-tcl')
def processor(env):
    proc = ProcessorStep('support-processor-tcl', name='processor', extra_env=env)
    proc.depends_on(link('tcl/processor_bd.tcl', proc.workdir))
    proc.creates('processor.dcp')
    proc.creates('.gen/sources_1/bd/processor/hdl/processor_wrapper.vhd')
    return proc


class BitbinStep(ProcessStep):
    def __init__(self, source, platform, **kwargs):
        self.arch = platform.arch
        self._workdir = source.path.parent
        super().__init__()
        self.depends_on(source)

        self.bitfile = source.path
        binfile = self.bitfile.with_suffix('.bit.bin')
        self.creates(Bitstream(binfile, platform=platform, **kwargs))

    @property
    def workdir(self):
        return self._workdir

    def command(self):
        bs = self.bitfile.name
        return f"${{BOOTGEN:-`which bootgen`}} -w on -image {bs}.bif -arch {self.arch} -process_bitstream bin"

    async def do_step(self):
        biffile = Path(self.bitfile).with_suffix('.bit.bif')
        biffile.write_text(f"all: {{ {self.bitfile.name} }}")
        await super().do_step()


class AbstractShell(VivadoStep):
    def __init__(self, platform, cell_name, pblock_name):
        self.platform = platform
        self.cell_name = cell_name
        self.pblock_name = pblock_name
        name = f'{platform.name}-{self.safe_cell_name}'
        shell_dcp = f'{name}.dcp'

        super().__init__(f'support-shell-{platform.arch}-tcl',
                         name=name,
                         dcp='platform_route.dcp',
                         tcl_args=(cell_name, shell_dcp))

        self.depends_on(link(platform.route(), self.workdir))
        self.creates(shell_dcp)

    @property
    def plat_name(self):
        return self.platform.name

    @property
    def path(self):
        return next(iter(self.outputs())).path

    @property
    def safe_cell_name(self):
        name = self.cell_name.translate(
            str.maketrans({"[": "_", "]": None, "*": "_"}))
        name = name.replace('_INSTR', '')
        return name

    @property
    def slot(self):
        """ This makes some big assumptions about cell naming conventions"""
        if (m := re.search(r'INSTRS\[(\d+)\]', self.cell_name)) is None:
            return None
        return int(m.group(1))

    @cache
    def sha(self):
        sha = hashlib.sha256(self.path.read_bytes())
        return sha.hexdigest()


class Platform(BuildStep):
    def __init__(self, env, ip, vhdl, slots, slot_ains, slot_aouts, slot_cbufs):
        self.slots = slots
        self.slot_ains = slot_ains
        self.slot_aouts = slot_aouts
        self.slot_cbufs = slot_cbufs
        self.env = env
        self.ip = ip
        self.vhdl = vhdl
        super().__init__(name=f'{self.hgboard}-{slots:d}')
        self._shells = {}

        self.depends_on(self.bitstreams())
        self.depends_on(Timing(self.reports(), self.name))

    @property
    def hgboard(self):
        return self.env['HG_BOARD']

    @property
    def part(self):
        # TODO remove this env stuff
        return self.env['HG_PART']

    @property
    def hwver(self):
        return self.env['HG_HWVER']

    @property
    def arch(self):
        return self.env['HG_ARCH']

    @property
    def top(self):
        return self.env['HG_TOP']

    @property
    def sha(self):
        return self._platform_bit.sha()

    def outputs(self):
        return self.inputs()

    def slot_shells(self):
        return sorted([v for v in self._shells.values() if v.slot is not None], key=lambda s: s.slot)

    @cache
    def create_shell(self, cell_name, pblock_name):
        if cell_name in self._shells:
            return self._shells[cell_name]

        shell = AbstractShell(self, cell_name, pblock_name)
        self._shells[cell_name] = shell
        return shell

    def shell(self, cell_name):
        return self._shells[cell_name]

    @cache  # multiple calls should not create new instances
    def synth(self):
        synth = SynthStep('support-synth-tcl',
                          name=f'synth-{self.slots:d}',
                          tcl_args=(self.slots, self.slot_ains, self.slot_aouts, self.slot_cbufs),
                          extra_env=self.env)
        synth.depends_on(link('processor', synth.workdir),
                         link('constraints/constraints.xdc', synth.workdir / 'constraints'),
                         link(self.vhdl, synth.workdir / 'vhdl'),
                         link('support-vhdl', synth.workdir / 'vhdl/support'),
                         link(self.ip, synth.workdir / 'ip'))
        synth.creates('platform_synth.dcp')
        global_alias('all-synth', synth)
        return synth

    @cache
    def route(self):
        synth = self.synth()
        route = RouteStep('support-route-tcl',
                          name=f'route-{self.slots:d}',
                          dcp='platform_synth.dcp')
        route.depends_on(link(synth, route.workdir),
                         link('constraints/constraints.xdc', route.workdir / 'constraints'),
                         link(f'constraints/floorplan_{self.slots:d}.xdc', route.workdir / 'constraints'),
                         link('support-slot-constraints', route.workdir))
        return route.creates('platform_route.dcp')

    @cache
    def bitstreams(self):
        route = self.route()
        bitstreams = VivadoStep('support-bitstreams-tcl',
                                name=f'bitstreams-{self.slots:d}',
                                dcp='platform_route.dcp')
        bitstreams.depends_on(link(route, bitstreams.workdir))
        self._platform_xsa = bitstreams.creates('platform.xsa')
        bitbin = BitbinStep(bitstreams.creates('platform.bit').pop(), platform=self)
        self._platform_bit = list(bitbin.outputs())[0]
        self._empty_bits = [
            BitbinStep(bitstreams.creates(
                f'platform_pblock_INSTR{i:d}_partial.bit').pop(),
                platform=self, slot=i)
            for i in range(self.slots)]
        return (self._platform_bit, self._empty_bits)

    @cache
    def reports(self):
        route = self.route()
        name = f'{self.hgboard}-{self.slots:d}'

        s = VivadoStep('support-platform-reports-tcl',
                       name=f'reports-{name}',
                       dcp='platform_route.dcp',
                       tcl_args=(name,))
        s.depends_on(link(route, s.workdir))
        s.creates(f'{name}_util.rpt',
                  f'{name}_timing.rpt',
                  f'{name}_util_hier.rpt')
        return s

    def platform_bit(self):
        return self._platform_bit

    def empty_bits(self):
        return self._empty_bits


class Bitstream(SourceFile):
    def __init__(self, path, *args, **kwargs):
        self.slot = kwargs.pop('slot', None)
        self.platform = kwargs.pop('platform')
        self.manifest_extra = kwargs.pop('manifest_extra', {})
        super().__init__(path, *args, **kwargs)

    @cache
    def sha(self):
        sha = hashlib.sha256(self.path.read_bytes())
        return sha.hexdigest()


class PythonZip(BuildStep):
    def __init__(self, outfile, **kwargs):
        super().__init__(**kwargs)
        self.path = next(iter(self.creates(outfile))).path
        self.platform = set()  # HACK
        self.depends_on('support-slot-python')

    async def do_step(self):
        with zipfile.ZipFile(self.path, mode='w') as f:
            # using PyZipFile causes issues across different python versions
            for inp in self.inputs():
                if inp.path.name == 'api_main.py':
                    f.write(inp.path, arcname='__main__.py')
                else:
                    f.write(inp.path, arcname='__pymoku_plugin/' + inp.path.name)  # TODO don't flatten everything

    @cache
    def sha(self):
        sha = hashlib.sha256(self.path.read_bytes())
        return sha.hexdigest()


class RMSynth(SynthStep):
    def __init__(self, top, vhdl, fpga_part, config_name=None, ip=[], generics={}, env={},
                 synth_tcl='support-slot-synth-tcl',
                 constraints='support-slot-constraints'):
        self.config_name = config_name
        self.generics = generics

        generics_str = ' '.join([f'{k}={str(v).lower()}' for k, v in generics.items()])
        generics_str = f'"{generics_str}"'
        super().__init__(synth_tcl,
                         name=f'{config_name}-synth',
                         extra_env=env,
                         tcl_args=(top, fpga_part, generics_str))

        for core in ip:
            self.depends_on(copy(core, self.workdir / 'ip' / core.path.stem))
        self.depends_on(link('support-slot-vhdl', self.workdir / 'vhdl/support')),
        self.depends_on(link(vhdl, self.workdir / 'vhdl'))
        self.depends_on(link(constraints, self.workdir / 'constraints'))
        self.creates('synth.dcp')
        global_alias('all-synth', self)


class RMRoute(RouteStep):
    def __init__(self, synth, shell,
                 route_tcl='support-slot-route-tcl',
                 constraints='support-slot-constraints'):
        self.config_name = f'{synth.config_name}-{shell.safe_cell_name}'
        self.shell = shell
        self.synth = synth
        super().__init__(route_tcl,
                         name=f'{self.config_name}-route',
                         dcp=shell.path.name,
                         tcl_args=(synth.workdir.absolute() / 'synth.dcp',
                                   shell.cell_name,
                                   'route.dcp'))
        self.depends_on(synth)
        self.depends_on(link(shell, self.workdir))
        self.depends_on(link(constraints, self.workdir / 'constraints'))
        self.creates('route.dcp')


class RMInstance(BuildStep):
    def __init__(self, route,
                 bitstream_tcl='support-slot-bitstream-tcl',
                 reports_tcl='support-slot-reports-tcl'):
        self.bitstream_tcl = bitstream_tcl
        self.reports_tcl = reports_tcl
        super().__init__()
        self.depends_on(self.reports(route))
        self.creates(self.bitstream(route).outputs())
        self.creates(self.reports(route).outputs())

    @cache
    def bitstream(self, route):
        b = VivadoStep(self.bitstream_tcl,
                       name=f'{route.config_name}-bitstream',
                       dcp='route.dcp',
                       tcl_args=(route.shell.cell_name, f'out_{route.shell.safe_cell_name}'))
        b.depends_on(link(route, b.workdir))
        bs = BitbinStep(
            b.creates(f'out_{route.shell.safe_cell_name}.bit').pop(),
            platform=route.shell.platform,
            slot=route.shell.slot,
            manifest_extra={
                k.lower(): v for k, v in route.synth.generics.items()
            })
        return bs

    @cache
    def reports(self, route):
        s = VivadoStep(self.reports_tcl,
                       name=f'{route.config_name}-reports',
                       dcp='route.dcp',
                       tcl_args=(route.shell.cell_name, route.shell.pblock_name))
        s.depends_on(link(route, s.workdir))
        s.creates('pblock_util.rpt', 'timing.rpt')

        name = f'{route.config_name}'
        if self.prefix:
            name = f'{self.prefix}-{name}'

        return Timing(s, name=name)


class Timing(BuildStep):
    results_table = None

    @dataclass
    class Stats(object):
        clb: tuple[int, int, float] = (0, 0, 0)
        dsp: tuple[int, int, float] = (0, 0, 0)
        bram: tuple[int, int, float] = (0, 0, 0)
        wns: float = 0

    def __init__(self, reports, name, **kwargs):
        super().__init__(**kwargs)
        self.row_name = name
        self.depends_on(reports)

    def state(self):
        return b''

    def current(self, in_state):
        return False

    async def do_step(self):
        if Timing.results_table is None and _with_rich:
            table = Table(show_edge=False, pad_edge=False, padding=0)
            Timing.results_table = table
            self.output_table.add_row(table)

            table.add_column(" ", justify="left", style="cyan", no_wrap=True)
            table.add_column("CLB", justify="center", width=17)
            table.add_column("DSP", justify="center", width=27)
            table.add_column("BRAM18", justify="center", width=27)
            table.add_column("WNS", justify="center", width=6)

        data = {}
        for inp in self.inputs():
            if inp.path.name.endswith('util.rpt'):
                data[self.row_name] = data.get(self.row_name, self.Stats())
                self.parse_util(inp.path, data[self.row_name])
            elif inp.path.name.endswith('timing.rpt'):
                data[self.row_name] = data.get(self.row_name, self.Stats())
                self.parse_timing(inp.path, data[self.row_name])
            else:
                continue

        for k, v in data.items():
            self.print_row(k, v)

    def print_row(self, name, stats):
        if _with_rich and (show_stats or stats.wns < 0.0):
            text = [Text(name),
                    Columns([Text(f'{stats.clb[2]:5.1f}%'),
                             Bar(100.0, 0.0, stats.clb[2], width=10, bgcolor='grey11', color='sandy_brown' if stats.clb[2] > 80.0 else 'default')]),
                    Columns([Text(f'{stats.dsp[2]:5.1f}%'),
                             Bar(100.0, 0.0, stats.dsp[2], width=10, bgcolor='grey11'),
                             Text(f'{stats.dsp[0]:4d}/{stats.dsp[1]:4d}')]),
                    Columns([Text(f'{stats.bram[2]:5.1f}%'),
                             Bar(100.0, 0.0, stats.bram[2], width=10, bgcolor='grey11'),
                             Text(f'{stats.bram[0]:4d}/{stats.bram[1]:4d}')]),
                    Text(f'{stats.wns:6.3f}', style='red' if stats.wns < 0.0 else None)]
            self.results_table.add_row(*text)

        else:
            self.log.info(f'{name}:'
                          f' CLB {stats.clb[2]:5.1f}%'
                          f' DSP {stats.dsp[2]:5.1f}%'
                          f' BRAM {stats.bram[2]:5.1f}%'
                          f' WNS {stats.wns:6.3f}')
            return

    def parse_util(self, path, stats):
        rpt = path.read_text()

        m = re.search(r'^\|\s(?:CLB|Slice)\s+\|.*\|$', rpt, flags=re.MULTILINE)
        row = [x.strip() for x in m.group(0).split('|')]
        used, _, _, avail, util = row[-6:-1]
        stats.clb = int(used), int(avail), float(util)

        m = re.search(r'^\|\sDSPs\s+\|.*\|$', rpt, flags=re.MULTILINE)
        row = [x.strip() for x in m.group(0).split('|')]
        used, _, _, avail, util = row[-6:-1]
        stats.dsp = int(used), int(avail), float(util)

        m = re.search(r'^\|\sBlock RAM Tile\s+\|.*\|$', rpt, flags=re.MULTILINE)
        row = [x.strip() for x in m.group(0).split('|')]
        used, _, _, avail, util = row[-6:-1]
        stats.bram = int(2 * float(used)), int(avail), float(util)

    def parse_timing(self, path, stats):
        rpt = path.read_text()
        # Slack (MET) :             0.140ns  (arrival time - required time)
        m = re.search(r'\(\bVIOLATED\b\)\s\:\s+([\-\d\.]+)ns', rpt, flags=re.MULTILINE)

        if m is not None:
            stats.wns = float(m.group(1))


# TODO ip is currently synth'd in each individual step that uses it which is
# a bit redundant.  There are some issues around vivado outputting encrypted
# netlists (.edn), as well as our ability to reuse a single XCI for multiple
# platforms
def gen_ip(xcis, platform):
    steps = []
    for xci in xcis:
        assert isinstance(xci, SourceFile)
        s = VivadoStep('support-synthip-tcl',
                       name=f'ip-{platform.hgboard}-{xci.path.stem}',
                       tcl_args=(xci.path.name,),
                       extra_env=platform.env)
        s.depends_on(link(xci, s.workdir))
        if (coe := xci.path.with_suffix('.coe')).exists():
            s.depends_on(link(coe, s.workdir))

        s.creates(f'{xci.path.stem}.dcp')
        steps.append(s)
    return steps



from pymoku.parts import DockerStep, link


class CocoTB(DockerStep):
    def __init__(self, name, top, sources, python, generics={}, *args, **kwargs):
        generics_str = ' '.join([f'{k}={str(v).lower()}' for k, v in generics.items()])

        extra_env = dict(
            TEST_TOP=top.lower(),
            TEST_MODULE=python,
            TEST_GENERICS=generics_str,
        )

        super().__init__('li-cocotb-pymoku',
                         name=name,
                         extra_env=extra_env,
                         **kwargs)
        self.depends_on(link('support-sim-python', self.workdir))
        self.depends_on(link('support-slot-vhdl', self.workdir / 'support-vhdl'))
        self.depends_on(link(sources, self.workdir))
        self.creates('wave.ghw')

    def command(self):
        return 'pytest -vs cocotb_test.py'


class CocoTBLib(DockerStep):
    def __init__(self, name, sources):
        super().__init__('li-cocotb-pymoku', name=name)
        self.depends_on(link(sources, self.workdir))
        self.creates(f'{self.name}-obj08.cf')
        self.creates(s.path.name for s in sources)
        self.sources = sources

    def command(self):
        cmd = f"ghdl -a --work={self.name} "
        cmd += "--std=08 --ieee=synopsys -frelaxed-rules -P/unisim "
        cmd += " ".join(s.path.name for s in self.sources)
        return cmd



from pymoku.parts import BuildStep
from pymoku.sdk import PythonZip, Bitstream, AbstractShell, SCM_VERSION, FIRMWARE_VERSION


class BaseManifest:
    def __init__(self, _id, *args, aliases=[], **kwargs):
        self._id = _id
        self.aliases = aliases
        self._bitstreams = self._resolve_bitstreams(*args)

    def _resolve_bitstreams(self, *args):
        res = set()
        for a in args:
            if isinstance(a, Bitstream):
                res.add(a)
            elif isinstance(a, BuildStep):
                res.update(self._resolve_bitstreams(*a.outputs()))
            elif hasattr(a, '__iter__'):
                res.update(self._resolve_bitstreams(*a))
        return res

    def bitstreams(self):
        return self._bitstreams

    def entry_for_item(self, item, arcname_for):
        return dict(
            _id=self._id,
            _type=self.TYPE,
            # aliases=self.aliases,  # TODO conflict with firmware
            sha256=item.sha(),
            version=SCM_VERSION,
            filename=arcname_for(item),
            size=item.path.stat().st_size,
        )

    def __call__(self, arcname_for):
        return [(self.entry_for_item(bs, arcname_for), bs) for bs in self.bitstreams()]


class Manifest(BaseManifest):
    TYPE = 'instrument'

    def entry_for_item(self, item, arcname_for):
        d = super().entry_for_item(item, arcname_for)
        d.update(
            hwver=item.platform.hwver,
            plat_sha=item.platform.sha,
            slot=item.slot,
            extra=item.manifest_extra,
        )
        return d


class PlatformManifest(BaseManifest):
    TYPE = 'platform'

    def entry_for_item(self, item, arcname_for):
        d = super().entry_for_item(item, arcname_for)
        d.update(
            hwver=item.platform.hwver,
            slots=item.platform.slots,
            fwcompat=FIRMWARE_VERSION,
            inputs=item.platform.slot_ains,
            outputs=item.platform.slot_aouts,
            slot_outputs=item.platform.slot_aouts,
            slot_cbufs=item.platform.slot_cbufs,
            loops=0  # COMPAT
        )
        return d


class PythonManifest(BaseManifest):
    TYPE = 'python'

    def _resolve_bitstreams(self, *args):
        res = set()
        for a in args:
            if isinstance(a, PythonZip):
                res.add(a)
            else:
                res.update(self._resolve_bitstreams(*a))
        return res


class NetworkManifest(BaseManifest):
    TYPE = 'network'

    def entry_for_item(self, item, arcname_for):
        d = super().entry_for_item(item, arcname_for)
        d.update(
            hwver=item.platform.hwver,
            plat_sha=item.platform.sha,
        )
        return d


class CheckpointManifest(BaseManifest):
    TYPE = 'checkpoint'

    def _resolve_bitstreams(self, *args):
        res = set()
        for a in args:
            if isinstance(a, AbstractShell):
                res.add(a)
            else:
                res.update(self._resolve_bitstreams(*a))
        return res

    def entry_for_item(self, item, arcname_for):
        d = super().entry_for_item(item, arcname_for)
        d.update(
            plat_sha=item.platform.sha,
            extra=dict(
                hwver=item.platform.hwver,
                part=item.platform.part,
                arch=item.platform.arch,
                cell_name=item.cell_name,
                pblock_name=item.pblock_name,
                slot=item.slot,
                # TODO try to keep platform abstraction?
                num_ain=item.platform.slot_ains,
                num_aout=item.platform.slot_aouts,
                num_cbufs=item.platform.slot_cbufs,
                # TODO
                # dsps. clbs, brams
            ),
        )
        return d



import os
import sys
import pytest
from pathlib import Path
from cocotb.runner import get_runner, Simulator

sim = os.getenv('SIM', 'ghdl')
toplevel = os.getenv('TEST_TOP')
sim_args = ["--wave=wave.ghw", "--ieee-asserts=disable"]
extra_args = ["--std=08", "--ieee=synopsys", "-frelaxed-rules", "-P/unisim", "-Pmoku"]
extra_env = {"COCOTB_RESOLVE_X": "ZEROS", "COCOTB_REDUCED_LOG_FMT": "1"}

generics = os.getenv('TEST_GENERICS')
if generics:
    build_generics = dict(x.split('=') for x in generics.split(' '))
else:
    build_generics = []


@pytest.fixture(scope="module")
def runner() -> Simulator:
    build_dir = '.'

    vhdl_sources = list(Path().glob('*.vhd'))
    vhdl_sources += list((Path('support-vhdl')).glob('*.vhd'))

    # lib_vhdl_sources = list((this_dir / 'lib/hdl/moku').glob('*.vhd'))

    runner = get_runner(sim)

    # runner.build(
    #     hdl_library='moku',
    #     vhdl_sources=lib_vhdl_sources,
    #     build_args=extra_args,
    #     build_dir=build_dir)

    runner.build(
        hdl_library='work',
        vhdl_sources=vhdl_sources,
        hdl_toplevel=toplevel,
        build_args=extra_args,
        parameters=build_generics,
        build_dir=build_dir)

    runner.test_args = extra_args

    return runner


def test_bench(runner: Simulator):
    runner.test(
        hdl_toplevel=toplevel,
        hdl_toplevel_library="work",
        test_module='startup',
        plusargs=sim_args,
        waves=True,
        extra_env=extra_env,
    )
